@@ -1,7 +1,7 @@
 #
 # SNMP Devices Simulator
 #
-# Written by Ilya Etingof <ilya@glas.net>, 2010-2011
+# Written by Ilya Etingof <ilya@glas.net>, 2010-2012
 #
 import os
 import sys
@@ -20,12 +20,21 @@ else:
 import bisect
 from pyasn1.type import univ
 from pyasn1.codec.ber import encoder, decoder
+from pyasn1.compat.octets import octs2str
 from pysnmp.entity import engine, config
 from pysnmp.entity.rfc3413 import cmdrsp, context
 from pysnmp.carrier.asynsock.dgram import udp
+try:
+    from pysnmp.carrier.asynsock.dgram import udp6
+except ImportError:
+    udp6 = None
+try:
+    from pysnmp.carrier.asynsock.dgram import unix
+except ImportError:
+    unix = None
 from pysnmp.carrier.asynsock.dispatch import AsynsockDispatcher
 from pysnmp.smi import exval, indices
-from pysnmp.proto import rfc1902, api
+from pysnmp.proto import rfc1902, rfc1905, api
 from pysnmp import debug
 
 # Process command-line options
@@ -37,17 +46,37 @@ v2cArch = False
 v3Only = False
 v3User = 'simulator'
 v3AuthKey = 'auctoritas'
+v3AuthProto = 'MD5'
 v3PrivKey = 'privatus'
-v3PrivProto = config.usmDESPrivProtocol
-agentAddress = ('127.0.0.1', 161)
-deviceDir = '.'
+v3PrivProto = 'DES'
+agentUDPv4Address = ('127.0.0.1', 161)
+agentUDPv4Endpoints = []
+agentUDPv6Endpoints = []
+agentUNIXEndpoints = []
+deviceDirs = set()
 deviceExt = os.path.extsep + 'snmpwalk'
 
-helpMessage = 'Usage: %s [--help] [--debug=<category>] [--device-dir=<dir>] [--force-index-rebuild] [--validate-device-data] [--agent-address=<X.X.X.X>] [--agent-port=<port>] [--v2c-arch ] [--v3-only] [--v3-user=<username>] [--v3-auth-key=<key>] [--v3-priv-key=<key>] [--v3-priv-proto=<DES|AES>]' % sys.argv[0]
+authProtocols = {
+  'MD5': config.usmHMACMD5AuthProtocol,
+  'SHA': config.usmHMACSHAAuthProtocol,
+  'NONE': config.usmNoAuthProtocol
+}
+
+privProtocols = {
+  'DES': config.usmDESPrivProtocol,
+  '3DES': config.usm3DESEDEPrivProtocol,
+  'AES': config.usmAesCfb128Protocol,
+  'AES128': config.usmAesCfb128Protocol,
+  'AES192': config.usmAesCfb192Protocol,
+  'AES256': config.usmAesCfb256Protocol,
+  'NONE': config.usmNoPrivProtocol
+}
+ 
+helpMessage = 'Usage: %s [--help] [--debug=<category>] [--device-dir=<dir>] [--force-index-rebuild] [--validate-device-data] [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>] [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>] [--agent-unix-endpoint=</path/to/named/pipe>] [--v2c-arch] [--v3-only] [--v3-user=<username>] [--v3-auth-key=<key>] [--v3-auth-proto=<%s>] [--v3-priv-key=<key>] [--v3-priv-proto=<%s>]' % (sys.argv[0], '|'.join(authProtocols), '|'.join(privProtocols))
 
 try:
     opts, params = getopt.getopt(sys.argv[1:], 'h',
-        ['help', 'debug=', 'device-dir=', 'force-index-rebuild', 'validate-device-data', 'agent-address=', 'agent-port=', 'v2c-arch', 'v3-only', 'v3-user=', 'v3-auth-key=', 'v3-priv-key=', 'v3-priv-proto=']
+        ['help', 'debug=', 'device-dir=', 'force-index-rebuild', 'validate-device-data', 'agent-address=', 'agent-port=', 'agent-udpv4-endpoint=', 'agent-udpv6-endpoint=', 'agent-unix-endpoint=', 'v2c-arch', 'v3-only', 'v3-user=', 'v3-auth-key=', 'v3-auth-proto=', 'v3-priv-key=', 'v3-priv-proto=']
         )
 except Exception:
     sys.stdout.write('%s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
@@ -64,15 +93,43 @@ for opt in opts:
     elif opt[0] == '--debug':
         debug.setLogger(debug.Debug(opt[1]))
     elif opt[0] == '--device-dir':
-        deviceDir = opt[1]
+        deviceDirs.add(opt[1])
     elif opt[0] == '--force-index-rebuild':
         forceIndexBuild = True
     elif opt[0] == '--validate-device-data':
         validateData = True
+    elif opt[0] == '--agent-udpv4-endpoint':
+        f = lambda h,p=161: (h, int(p))
+        try:
+            agentUDPv4Endpoints.append(f(*opt[1].split(':')))
+        except:
+            sys.stdout.write('improper IPv4/UDP endpoint %s\r\n' % opt[1])
+            sys.exit(-1)
+    elif opt[0] == '--agent-udpv6-endpoint':
+        if not udp6:
+            sys.stdout.write('This system does not support UDP/IP6\r\n')
+            sys.exit(-1)
+        if opt[1].find(']:') != -1 and opt[1][0] == '[':
+            h, p = opt[1].split(']:')
+            try:
+                h, p = h[1:], int(p)
+            except:
+                sys.stdout.write('improper IPv6/UDP endpoint %s\r\n' % opt[1])
+                sys.exit(-1)
+        elif opt[1][0] == '[' and opt[1][-1] == ']':
+            h, p = opt[1][1:-1], 161
+        else:
+            h, p = opt[1], 161
+        agentUDPv6Endpoints.append((h, p))
+    elif opt[0] == '--agent-unix-endpoint':
+        if not unix:
+            sys.stdout.write('This system does not support UNIX domain sockets\r\n')
+            sys.exit(-1)
+        agentUNIXEndpoints.append(opt[1])
     elif opt[0] == '--agent-address':
-        agentAddress = (opt[1], agentAddress[1])
+        agentUDPv4Address = (opt[1], agentUDPv4Address[1])
     elif opt[0] == '--agent-port':
-        agentAddress = (agentAddress[0], int(opt[1]))
+        agentUDPv4Address = (agentUDPv4Address[0], int(opt[1]))
     elif opt[0] == '--v2c-arch':
         v2cArch = True
     elif opt[0] == '--v3-only':
@@ -81,17 +138,33 @@ for opt in opts:
         v3User = opt[1]
     elif opt[0] == '--v3-auth-key':
         v3AuthKey = opt[1]
+    elif opt[0] == '--v3-auth-proto':
+        v3AuthProto = opt[1].upper()
+        if v3AuthProto not in authProtocols:
+            sys.stdout.write('bad v3 auth protocol %s\r\n' % v3AuthProto)
+            sys.exit(-1)
     elif opt[0] == '--v3-priv-key':
         v3PrivKey = opt[1]
     elif opt[0] == '--v3-priv-proto':
-        if opt[1].upper() == 'DES':
-            v3PrivProto = config.usmDESPrivProtocol
-        elif opt[1].upper() == 'AES':
-            v3PrivProto = config.usmAesCfb128Protocol
-        else:
-            sys.stdout.write('bad v3 privacy protocol\r\n')
+        v3PrivProto = opt[1].upper()
+        if v3PrivProto not in privProtocols:
+            sys.stdout.write('bad v3 privacy protocol %s\r\n' % v3PrivProto)
             sys.exit(-1)
 
+if authProtocols[v3AuthProto] == config.usmNoAuthProtocol and \
+    privProtocols[v3PrivProto] != config.usmNoPrivProtocol:
+        sys.stdout.write('privacy impossible without authentication\r\n')
+        sys.exit(-1)
+
+if not deviceDirs:
+    deviceDirs.add('devices')
+
+# for backward compatibility
+if not agentUDPv4Endpoints and \
+   not agentUDPv6Endpoints and \
+   not agentUNIXEndpoints:
+    agentUDPv4Endpoints.append(agentUDPv4Address)
+
 # Device file entry parsers
 
 class DumpParser:
@@ -124,7 +197,7 @@ class DumpParser:
         '6': __unhexFilter
         }
 
-    def parse(self, line): return line.split('|', 2)
+    def parse(self, line): return octs2str(line).split('|', 2)
 
     def evaluateOid(self, oid):
         return univ.ObjectIdentifier(oid)
@@ -168,7 +241,7 @@ class SnmprecParser:
                rfc1902.Counter64 ):
         tagMap[str(sum([ x for x in t.tagSet[0] ]))] = t
 
-    def parse(self, line): return line.strip().split('|', 2)
+    def parse(self, line): return octs2str(line).strip().split('|', 2)
 
     def evaluateOid(self, oid):
         return univ.ObjectIdentifier(oid)
@@ -234,6 +307,9 @@ class DeviceFile:
                 if textFileStamp < os.stat(dbFile)[8]:
                     if indexNeeded:
                         sys.stdout.write('Forced index rebuild %s\r\n' % dbFile)
+                    elif not whichdb(dbFile):
+                        indexNeeded = True
+                        sys.stdout.write('Unsupported index format, rebuilding index %s\r\n' % dbFile)
                 else:
                     indexNeeded = True
                     sys.stdout.write('Index %s out of date\r\n' % dbFile)
@@ -243,12 +319,22 @@ class DeviceFile:
             sys.stdout.write('Index does not exist for %s\r\n' % self.__textFile)
             
         if indexNeeded:
-            sys.stdout.write('Indexing device file %s...' % self.__textFile)
-            sys.stdout.flush()
+            # these might speed-up indexing
+            open_flags = 'nfu' 
+            while open_flags:
+                try:
+                    db = dbm.open(self.__dbFile, open_flags)
+                except:
+                    open_flags = open_flags[:-1]
+                    if not open_flags:
+                        raise
+                else:
+                    break
 
-            text = open(self.__textFile)
-            
-            db = dbm.open(self.__dbFile, 'n')
+            text = open(self.__textFile, 'rb')
+
+            sys.stdout.write('Indexing device file %s (open flags \"%s\")...' % (self.__textFile, open_flags))
+            sys.stdout.flush()
         
             lineNo = 0
             offset = 0
@@ -257,7 +343,7 @@ class DeviceFile:
                 if not line:
                     break
             
-                lineNo = lineNo + 1
+                lineNo += 1
 
                 try:
                     oid, tag, val = textParser.parse(line)
@@ -291,7 +377,7 @@ class DeviceFile:
                         
                 db[oid] = str(offset)
 
-                offset = text.tell()
+                offset += len(line)
 
             text.close()
             db.close()
@@ -315,7 +401,7 @@ class DeviceFile:
 
             DeviceFile.openedQueue.append(self)
 
-            self.__text = open(self.__textFile)
+            self.__text = open(self.__textFile, 'rb')
             
             self.__db = dbm.open(self.__dbFile)
 
@@ -365,20 +451,23 @@ class MibInstrumController:
     # In-place, by-OID binary search
 
     def __searchOid(self, text, oid):
-        lo = mid = 0
+        lo = mid = 0; prev_mid = -1
         text.seek(0, 2)
         hi = sz = text.tell()
         while lo < hi:
             mid = (lo+hi)//2
+            text.seek(mid)
             while mid:
-                text.seek(mid)
                 c = text.read(1)
                 if c == '\n':
                     mid = mid + 1
                     break
-                mid = mid - 1
-            if not mid:
+                mid = mid - 1    # pivot stepping back in search for full line
                 text.seek(mid)
+            if mid == prev_mid:  # loop condition due to stepping back pivot
+                break
+            else:
+                prev_mid = mid
             line = text.readline()
             try:
                 midval, _ = self.__textParser.evaluate(line, oidOnly=True)
@@ -410,9 +499,9 @@ class MibInstrumController:
         text, db = self.__deviceFile.getHandles()
         
         for oid, val in varBinds:
-            textOid = univ.OctetString(
-                '.'.join([ '%s' % x for x in oid ])
-                ).asOctets()
+            textOid = str(
+                univ.OctetString('.'.join([ '%s' % x for x in oid ]))
+                )
             try:
                 offset = db[textOid]
                 exactMatch = True
@@ -478,7 +567,7 @@ class DevicesIndexInstrumController:
             return key, self.__db[key]
                                                             
     def readNextVars(self, varBinds, acInfo=None):
-        return [ (vb[0], self.__getNextVal(vb[0], exval.endOfMib)) for vb in varBinds ]        
+        return [ self.__getNextVal(vb[0], exval.endOfMib) for vb in varBinds ]        
 
     def writeVars(self, varBinds, acInfo=None):
         return [ (vb[0], exval.noSuchInstance) for vb in varBinds ]        
@@ -495,7 +584,7 @@ devicesIndexInstrumController = DevicesIndexInstrumController()
 # Basic SNMP engine configuration
 
 if v2cArch:
-    contexts = {}
+    contexts = { univ.OctetString('index'): devicesIndexInstrumController }
 else:
     snmpEngine = engine.SnmpEngine()
 
@@ -504,44 +593,50 @@ else:
     snmpContext = context.SnmpContext(snmpEngine)
         
     config.addV3User(
-        snmpEngine, v3User,
-        config.usmHMACMD5AuthProtocol, v3AuthKey,
-        v3PrivProto, v3PrivKey
+        snmpEngine,
+        v3User,
+        authProtocols[v3AuthProto], v3AuthKey,
+        privProtocols[v3PrivProto], v3PrivKey
         )
-        
+
 # Build pysnmp Managed Objects base from device files information
 
-for fullPath, textParser, communityName in getDevices(deviceDir):
-    mibInstrum = MibInstrumController(
-        DeviceFile(fullPath).indexText(textParser, forceIndexBuild), textParser
+for deviceDir in deviceDirs:
+    sys.stdout.write(
+        'Using data directory "%s"\r\n%s\r\n' % (deviceDir, '='*66)
+    )
+    for fullPath, textParser, communityName in getDevices(deviceDir):
+        mibInstrum = MibInstrumController(
+            DeviceFile(fullPath).indexText(textParser,
+            forceIndexBuild), textParser
         )
 
-    sys.stdout.write('Device %s\r\nSNMPv1/2c community name: %s\r\n' % \
-                     (mibInstrum, communityName))
+        sys.stdout.write('Device %s\r\nSNMPv1/2c community name: %s\r\n' % \
+                         (mibInstrum, communityName))
 
-    if v2cArch:
-        contexts[univ.OctetString(communityName)] = mibInstrum
+        if v2cArch:
+            contexts[univ.OctetString(communityName)] = mibInstrum
         
-        devicesIndexInstrumController.addDevice(
-            fullPath, communityName
+            devicesIndexInstrumController.addDevice(
+                fullPath, communityName
             )
-    else:
-        agentName = contextName = md5(univ.OctetString(communityName).asOctets()).hexdigest()
+        else:
+            agentName = contextName = md5(univ.OctetString(communityName).asOctets()).hexdigest()
 
-        if not v3Only:
-            config.addV1System(
-                snmpEngine, agentName, communityName, contextName=contextName
+            if not v3Only:
+                config.addV1System(
+                    snmpEngine, agentName, communityName, contextName=contextName
                 )
 
-        snmpContext.registerContextName(contextName, mibInstrum)
+            snmpContext.registerContextName(contextName, mibInstrum)
                  
-        devicesIndexInstrumController.addDevice(
-            fullPath, communityName, contextName
+            devicesIndexInstrumController.addDevice(
+                fullPath, communityName, contextName
             )
                  
-        sys.stdout.write('SNMPv3 context name: %s\r\n' % (contextName,))
+            sys.stdout.write('SNMPv3 context name: %s\r\n' % (contextName,))
         
-    sys.stdout.write('%s\r\n' % ('-+' * 33,))
+        sys.stdout.write('%s\r\n' % ('-+' * 33,))
         
 if v2cArch:
     def getBulkHandler(varBinds, nonRepeaters, maxRepetitions, readNextVars):
@@ -562,7 +657,30 @@ if v2cArch:
                     varBind = rspVarBinds[-1]
 
         return rspVarBinds
-    
+ 
+    # Suggest variations of context name based on request data
+
+    def probeContext(transportDomain, transportAddress, community):
+        candidate = [
+            community, '.'.join([ str(x) for x in transportDomain ])
+        ]
+        if transportDomain[:len(udp.domainName)] == udp.domainName:
+            candidate.append(transportAddress[0])
+        elif udp6 and transportDomain[:len(udp6.domainName)] == udp6.domainName:
+            candidate.append(
+                str(transportAddress[0]).replace(':', '_')
+            )
+        elif unix and transportDomain[:len(unix.domainName)] == unix.domainName:
+            candidate.append(transportAddress)
+
+        candidate = [ str(x) for x in candidate ]
+
+        while candidate:
+            yield rfc1902.OctetString(
+                      os.path.normpath(os.path.sep.join(candidate))
+                  )
+            del candidate[-1]
+
     def commandResponderCbFun(transportDispatcher, transportDomain,
                               transportAddress, wholeMsg):
         while wholeMsg:
@@ -575,12 +693,12 @@ if v2cArch:
             reqMsg, wholeMsg = decoder.decode(
                 wholeMsg, asn1Spec=pMod.Message(),
                 )
-            
+
             communityName = reqMsg.getComponentByPosition(1)
-            if communityName not in contexts:
-                sys.stdout.write('Unknown community name %s from %s:%s\r\n' % (
-                    communityName, transportDomain, transportAddress
-                    ))
+            for communityName in probeContext(transportDomain, transportAddress, communityName):
+                if communityName in contexts:
+                    break
+            else:
                 return wholeMsg
             
             rspMsg = pMod.apiMessage.getResponse(reqMsg)
@@ -616,14 +734,17 @@ if v2cArch:
                 )
 
             # Poor man's v2c->v1 translation
-            
+            errorMap = {  rfc1902.Counter64.tagSet: 5,
+                          rfc1905.NoSuchObject.tagSet: 2,
+                          rfc1905.NoSuchInstance.tagSet: 2,
+                          rfc1905.EndOfMibView.tagSet: 2  }
+ 
             if not msgVer:
                 for idx in range(len(varBinds)):
                     oid, val = varBinds[idx]
-                    if val.tagSet in (rfc1902.Counter64.tagSet,
-                                      univ.Null.tagSet):
+                    if val.tagSet in errorMap:
                         varBinds = pMod.apiPDU.getVarBinds(reqPDU)
-                        pMod.apiPDU.setErrorStatus(rspPDU, 5)
+                        pMod.apiPDU.setErrorStatus(rspPDU, errorMap[val.tagSet])
                         pMod.apiPDU.setErrorIndex(rspPDU, idx+1)
                         break
 
@@ -640,14 +761,35 @@ if v2cArch:
     contexts['index'] = devicesIndexInstrumController
     
     # Configure socket server
-    
+   
+    sys.stdout.write('Listening at:\r\n')
+ 
     transportDispatcher = AsynsockDispatcher()
-    transportDispatcher.registerTransport(
-            udp.domainName, udp.UdpTransport().openServerMode(agentAddress)
+    for idx in range(len(agentUDPv4Endpoints)):
+        transportDispatcher.registerTransport(
+                udp.domainName + (idx,),
+                udp.UdpTransport().openServerMode(agentUDPv4Endpoints[idx])
             )
+        sys.stdout.write('  UDP/IPv4 endpoint %s:%s, transport ID %s\r\n' % (agentUDPv4Endpoints[idx] + ('.'.join([str(x) for x in udp.domainName + (idx,)]),)))
+    for idx in range(len(agentUDPv6Endpoints)):
+        transportDispatcher.registerTransport(
+                udp6.domainName + (idx,),
+                udp6.Udp6Transport().openServerMode(agentUDPv6Endpoints[idx])
+            )
+        sys.stdout.write('  UDP/IPv6 endpoint %s:%s, transport ID %s\r\n' % (agentUDPv6Endpoints[idx] + ('.'.join([str(x) for x in udp6.domainName + (idx,)]),)))
+    for idx in range(len(agentUNIXEndpoints)):
+        transportDispatcher.registerTransport(
+                unix.domainName + (idx,),
+                unix.UnixTransport().openServerMode(agentUNIXEndpoints[idx])
+            )
+        sys.stdout.write('  UNIX domain endpoint %s, transport ID %s\r\n' % (agentUNIXEndpoints[idx], '.'.join([str(x) for x in unix.domainName + (idx,)])))
     transportDispatcher.registerRecvCbFun(commandResponderCbFun)
 else:
-    sys.stdout.write('SNMPv3 credentials:\r\nUsername: %s\r\nAuthentication key: %s\r\nEncryption (privacy) key: %s\r\nEncryption protocol: %s\r\n' % (v3User, v3AuthKey, v3PrivKey, v3PrivProto))
+    sys.stdout.write('SNMPv3 credentials:\r\nUsername: %s\r\n' % v3User)
+    if authProtocols[v3AuthProto] != config.usmNoAuthProtocol:
+        sys.stdout.write('Authentication key: %s\r\nAuthentication protocol: %s\r\n' % (v3AuthKey, v3AuthProto))
+        if privProtocols[v3PrivProto] != config.usmNoPrivProtocol:
+            sys.stdout.write('Encryption (privacy) key: %s\r\nEncryption protocol: %s\r\n' % (v3PrivKey, v3PrivProto))
 
     # Configure access to devices index
 
@@ -660,11 +802,29 @@ else:
 
     # Configure socket server
 
-    config.addSocketTransport(
-        snmpEngine,
-        udp.domainName,
-        udp.UdpTransport().openServerMode(agentAddress)
+    sys.stdout.write('Listening at:\r\n')
+
+    for idx in range(len(agentUDPv4Endpoints)):
+        config.addSocketTransport(
+            snmpEngine,
+            udp.domainName + (idx,),
+            udp.UdpTransport().openServerMode(agentUDPv4Endpoints[idx])
         )
+        sys.stdout.write('  UDP/IPv4 endpoint %s:%s, transport ID %s\r\n' % (agentUDPv4Endpoints[idx] + ('.'.join([str(x) for x in udp.domainName + (idx,)]),)))
+    for idx in range(len(agentUDPv6Endpoints)):
+        config.addSocketTransport(
+            snmpEngine,
+            udp6.domainName + (idx,),
+            udp6.Udp6Transport().openServerMode(agentUDPv6Endpoints[idx])
+        )
+        sys.stdout.write('  UDP/IPv6 endpoint %s:%s, transport ID %s\r\n' % (agentUDPv6Endpoints[idx] + ('.'.join([str(x) for x in udp6.domainName + (idx,)]),)))
+    for idx in range(len(agentUNIXEndpoints)):
+        config.addSocketTransport(
+            snmpEngine,
+            unix.domainName + (idx,),
+            unix.UnixTransport().openServerMode(agentUNIXEndpoints[idx])
+        )
+        sys.stdout.write('  UNIX domain endpoint %s, transport ID %s\r\n' % (agentUNIXEndpoints[idx], '.'.join([str(x) for x in unix.domainName + (idx,)])))
 
     # SNMP applications
 
@@ -674,10 +834,23 @@ else:
     cmdrsp.BulkCommandResponder(snmpEngine, snmpContext)
 
     transportDispatcher = snmpEngine.transportDispatcher
-    
-sys.stdout.write('\r\nListening at %s\r\n' % (agentAddress,))
 
 # Run mainloop
 
 transportDispatcher.jobStarted(1) # server job would never finish
-transportDispatcher.runDispatcher()
+
+# Python 2.4 does not support the "finally" clause
+
+exc_info = None
+
+try:
+    transportDispatcher.runDispatcher()
+except KeyboardInterrupt:
+    sys.stdout.write('Process terminated\r\n')
+except Exception:
+    exc_info = sys.exc_info()
+
+transportDispatcher.closeDispatcher()
+
+if exc_info:
+    raise exc_info[1]
